/*
 * main.cu
 *
 *  Created on: 2 mai 2013
 *      Author: technic105
 */

#include "main.cu.h"

//local app
#include "addon.cu.h"

int main(int argc, char* argv[])
{

	float *fpHostBuffer = NULL;
	float *fpDeviceBuffer = NULL;
	int size = 512;
	cudaStream_t  stream;
	dim3 blocks(size/512,1,1);
	dim3 threads(512,1,1);

	//Set cuda device flags regarding sync behaviour
	//cudaSetDeviceFlags ( cudaDeviceScheduleSpin ) ;

	//page-locked memory
	checkCudaErrors(cudaMallocHost(&fpHostBuffer,size*sizeof(float)));

	//Device memory
	checkCudaErrors(cudaMalloc(&fpDeviceBuffer,size*sizeof(float)));

	checkCudaErrors(cudaStreamCreate(&stream));

	/************************************************
	* Managing the function pointer buffer
	***********************************************/
	func* h_f;
	func* d_f;
	h_f = (func*)malloc(NUMBER_FUNCTION*sizeof(func));
	checkCudaErrors(cudaMalloc((void**)&d_f,NUMBER_FUNCTION*sizeof(func)));
	get_func_address(h_f);


	checkCudaErrors(cudaMemcpy(d_f,h_f,NUMBER_FUNCTION*sizeof(func),cudaMemcpyHostToDevice));
	/************************************************
	 * End of function pointer buffer management
	 ***********************************************/
	/************************************************
	* Managing the argument structures
	***********************************************/
	//Will contain the pointer to various type of argument struct in device memory
	funcParam** d_param;
	funcParam** d_param2;

	d_param =(funcParam**) malloc(NUMBER_FUNCTION*sizeof(funcParam*));

	//an instance of a specific argument struct
	funcParam1 fptest1;
	funcParam2 fptest2;
	funcParam3 fptest3;
	funcParam4 fptest4;
	fptest1.fpDeviceBuffer=fpDeviceBuffer;
	fptest2.fpDeviceBuffer=fpDeviceBuffer;
	fptest3.fpDeviceBuffer=fpDeviceBuffer;
	fptest4.fpDeviceBuffer=fpDeviceBuffer;
	//La première case de d_param contient l'adresse en device d'une zone pouvant contenir un funcparam1
	checkCudaErrors(cudaMalloc((void**)&(d_param[0]),sizeof(funcParam1)));
	checkCudaErrors(cudaMalloc((void**)&(d_param[1]),sizeof(funcParam2)));
	checkCudaErrors(cudaMalloc((void**)&(d_param[2]),sizeof(funcParam3)));
	checkCudaErrors(cudaMalloc((void**)&(d_param[3]),sizeof(funcParam4)));

	//On copie le funcparam1 dans l'adresse réservée à la ligne précédent
	checkCudaErrors(cudaMemcpy(d_param[0],&fptest1,sizeof(funcParam1),cudaMemcpyHostToDevice));
	checkCudaErrors(cudaMemcpy(d_param[1],&fptest2,sizeof(funcParam2),cudaMemcpyHostToDevice));
	checkCudaErrors(cudaMemcpy(d_param[2],&fptest2,sizeof(funcParam3),cudaMemcpyHostToDevice));
	checkCudaErrors(cudaMemcpy(d_param[3],&fptest2,sizeof(funcParam4),cudaMemcpyHostToDevice));


	//On réserve un espace en mémoire device pour copier le tableau de pointeurs vers les structs
	checkCudaErrors(cudaMalloc((void**)&(d_param2),NUMBER_FUNCTION*sizeof(funcParam*)));

	//On copie le tableau host contenant les pointeur de struct device vers le device
	checkCudaErrors(cudaMemcpy(d_param2,d_param,NUMBER_FUNCTION*sizeof(funcParam*),cudaMemcpyHostToDevice));

	/************************************************
	* Managing the argument structures
	***********************************************/

	checkCudaErrors( cudaMemcpyAsync ( fpDeviceBuffer, fpHostBuffer, size*sizeof(float), cudaMemcpyHostToDevice,stream) );
	calcbis<<<blocks,threads,0,stream>>>(fpDeviceBuffer,d_f,d_param2,NUMBER_FUNCTION);
	checkCudaErrors( cudaMemcpyAsync ( fpHostBuffer, fpDeviceBuffer, size*sizeof(float), cudaMemcpyDeviceToHost,stream) );
	checkCudaErrors( cudaStreamSynchronize(stream));

	checkCudaErrors( cudaMemcpyAsync ( fpDeviceBuffer, fpHostBuffer, size*sizeof(float), cudaMemcpyHostToDevice,stream) );
	calc1<<<blocks,threads,0,stream>>>(fptest1);
	//checkCudaErrors( cudaMemcpyAsync ( fpHostBuffer, fpDeviceBuffer, size*sizeof(float), cudaMemcpyDeviceToHost,stream) );
	//checkCudaErrors( cudaMemcpyAsync ( fpDeviceBuffer, fpHostBuffer, size*sizeof(float), cudaMemcpyHostToDevice,stream) );
	calc2<<<blocks,threads,0,stream>>>(fptest2);
	//checkCudaErrors( cudaMemcpyAsync ( fpHostBuffer, fpDeviceBuffer, size*sizeof(float), cudaMemcpyDeviceToHost,stream) );
	//checkCudaErrors( cudaMemcpyAsync ( fpDeviceBuffer, fpHostBuffer, size*sizeof(float), cudaMemcpyHostToDevice,stream) );
	calc3<<<blocks,threads,0,stream>>>(fptest3);
	//checkCudaErrors( cudaMemcpyAsync ( fpHostBuffer, fpDeviceBuffer, size*sizeof(float), cudaMemcpyDeviceToHost,stream) );
	//checkCudaErrors( cudaMemcpyAsync ( fpDeviceBuffer, fpHostBuffer, size*sizeof(float), cudaMemcpyHostToDevice,stream) );
	calc4<<<blocks,threads,0,stream>>>(fptest4);
	checkCudaErrors( cudaMemcpyAsync ( fpHostBuffer, fpDeviceBuffer, size*sizeof(float), cudaMemcpyDeviceToHost,stream) );
	checkCudaErrors( cudaStreamSynchronize(stream));


	checkCudaErrors(cudaStreamDestroy(stream));




	checkCudaErrors(cudaFree(d_f));
	for(int i = 0; i<NUMBER_FUNCTION;i++)
	{
		checkCudaErrors(cudaFree(d_param[i]));
	}
	free(h_f);
	free(d_param);
	checkCudaErrors(cudaFree(d_param2));
	checkCudaErrors(cudaFree(fpDeviceBuffer));
	checkCudaErrors(cudaFreeHost(fpHostBuffer));
	checkCudaErrors(cudaDeviceReset());

	return(EXIT_SUCCESS);
}


__global__ void calcbis(float* fpDeviceBuffer, func* f, funcParam** param, int n)
{
	//float x=fpDeviceBuffer[threadIdx.x+512*blockIdx.x];

	for(int i=0;i<n;++i)
	{
		//f[i](param[i]);
		//f[i](NULL);
		printf("adresse: %p\n",f[i]);
	}
}

__global__ void calc1(funcParam1 param)
{
	int pos=threadIdx.x+512*blockIdx.x;
	float x=param.fpDeviceBuffer[pos];
	//param.fpDeviceBuffer[pos]=(param.param3+(x+param.param0/param.param1))*param.param2;
	param.fpDeviceBuffer[pos]=x+param.param3;
}
__global__ void calc2(funcParam2 param)
{
	int pos=threadIdx.x+512*blockIdx.x;
	float x=param.fpDeviceBuffer[pos];
	//param.fpDeviceBuffer[pos]=(param.param3+(x+param.param0/param.param1))*param.param2;
	param.fpDeviceBuffer[pos]=x+param.param3;

}
__global__ void calc3(funcParam3 param)
{
	int pos=threadIdx.x+512*blockIdx.x;
	float x=param.fpDeviceBuffer[pos];
	//param.fpDeviceBuffer[pos]=(param.param3+(x+param.param0/param.param1))*param.param2;
	param.fpDeviceBuffer[pos]=x+param.param3;

}
__global__ void calc4(funcParam4 param)
{
	int pos=threadIdx.x+512*blockIdx.x;
	float x=param.fpDeviceBuffer[pos];
	//param.fpDeviceBuffer[pos]=(param.param3+(x+param.param0/param.param1))*param.param2;
	param.fpDeviceBuffer[pos]=x+param.param3;

}
